;; Generated package description from project-store.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "project-store" "0.9.0.0.20260819.2" "Project backend for Nix store" '((emacs "29.1")) :commit "6c2bb36826d0aa41e3842b44b24e6a649dbb386d" :authors '(("Lin Jian" . "me@linj.tech")) :maintainer '("Lin Jian" . "me@linj.tech") :keywords '("nix" "store" "project" "tools") :url "https://github.com/jian-lin/project-store")
